<?php

namespace Drupal\tfa_duo\Plugin\TfaValidation;

use Drupal\Component\Utility\HTML;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\encrypt\EncryptionProfileManagerInterface;
use Drupal\encrypt\EncryptServiceInterface;
use Drupal\tfa\Plugin\TfaBasePlugin;
use Drupal\tfa\Plugin\TfaValidationInterface;
use Drupal\user\Entity\User;
use Drupal\user\UserDataInterface;
use Duo\Web;

/**
 * Duo validation class for performing Duo validation.
 *
 * @TfaValidation(
 *   id = "tfa_duo",
 *   label = @Translation("Tfa Duo validation"),
 *   description = @Translation("Tfa Duo Validation Plugin"),
 *   isFallback = FALSE
 * )
 */
class TfaDuoValidation extends TfaBasePlugin implements TfaValidationInterface {
  use MessengerTrait;
  use StringTranslationTrait;

  /**
   * Object containing the external validation library.
   *
   * @var \Duo\Web
   */
  protected $duo;

  /**
   * This plugin's settings.
   *
   * @var array
   */
  protected $settings = [];

  /**
   * The KeyRepository.
   *
   * @var \Drupal\key\KeyRepository
   */
  protected $keyRepository;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, UserDataInterface $user_data, EncryptionProfileManagerInterface $encryption_profile_manager, EncryptServiceInterface $encrypt_service) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $user_data, $encryption_profile_manager, $encrypt_service);

    $this->duo = new Web();
    $plugin_settings = \Drupal::config('tfa.settings')->get('validation_plugin_settings');
    $this->settings = !empty($plugin_settings['tfa_duo']) ? $plugin_settings['tfa_duo'] : [];
    $this->keyRepository = \Drupal::service('key.repository');
  }

  /**
   * {@inheritdoc}
   */
  public function ready() {
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function getForm(array $form, FormStateInterface $form_state) {
    $key = $this->keyRepository->getKey($this->settings['duo_key'])->getKeyValues();

    // Verify the response from Duo if present in the request.
    if (!empty($_POST['sig_response'])) {
      $username = $this->duoVerifiedLogin($key['duo_integration'], $key['duo_secret'], $key['duo_application'], $_POST['sig_response']);
    }

    // If a valid username was returned, the user completed TFA and we
    // present them with the final form to login and auto-submit it.
    if (!empty($username) && user_load_by_name($username)) {
      return $this->getCompleteLoginForm($form, $form_state);
    }

    // Return the DUO iFrame form to present TFA options.
    return $this->getDuoIframe($form, $form_state);
  }

  /**
   * Returns the complete login form once a user has completed DUO TFA.
   *
   * @param array $form
   *   The form.
   * @param FormStateInterface $form_state
   *   The form state.
   * @return array
   *   The form elements.
   */
  protected function getCompleteLoginForm($form, FormStateInterface $form_state) {
    $key = $this->keyRepository->getKey($this->settings['duo_key'])->getKeyValues();
    $form['actions']['#type'] = 'actions';
    $url = Url::fromRoute('user.page');
    $form_state->setRedirect($url);

    // Store the response from Duo in hidden input fields for validation.
    foreach (['duo_integration', 'duo_secret', 'duo_application'] as $prop) {
      $form[$prop] = [
        '#type' => 'hidden',
        '#value' => $key[$prop],
      ];
    }

    $form['sig_response'] = [
      '#type' => 'hidden',
      '#value' => trim($_POST['sig_response']),
    ];

    $form['message'] = [
      '#markup' => $this->t('Duo authentication complete. Please wait while you are logged in. If you are not automatically redirected you can press the complete login button below.'),
    ];

    // Render a hidden button that's required to submit the form.
    $form['actions']['#type'] = 'actions';
    $form['actions']['login'] = [
      '#type' => 'submit',
      '#button_type' => 'primary',
      '#value' => $this->t('Complete login'),
      '#attributes' => [
        'class' => ['element-invisible'],
      ],
    ];

    // Attach the library to auto-submit the form and log the user in.
    $form['#attached'] = [
      'library' => [
        'tfa_duo/duo-auto-submit',
      ],
    ];

    return $form;
  }

  /**
   * Returns the DUO TFA iFrame form.
   *
   * @param array $form
   *   The base form.
   * @param FormStateInterface $form_state
   *   The form state.
   * @return array
   *   The DUO TFA iFrame form.
   */
  protected function getDuoIframe($form, FormStateInterface $form_state) {
    $key = $this->keyRepository->getKey($this->settings['duo_key'])->getKeyValues();

    // If there was no response from Duo, we'll build the Duo Iframe form.
    $sign_request = $this->duo->signRequest($key['duo_integration'], $key['duo_secret'], $key['duo_application'], $form_state->getBuildInfo()['args'][0]->getDisplayName());

    if ($this->responseErrors($sign_request)) {
      return $form;
    }

    $form['#title'] = t('Duo 2nd-Factor Login');
    $form['duo'] = [
      '#type' => 'inline_template',
      '#template' => '<iframe id="duo_iframe" data-host="{{ host }}" data-sig-request="{{ sign_request }}"></iframe>',
      '#context' => [
        'host' => $key['duo_apihostname'],
        'sign_request' => $sign_request,
      ],
    ];

    $form['#attached'] = [
      'library' => [
        'tfa_duo/duo-sign',
      ],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm($config, $state) {
    $form = [];
    $form['duo_key'] = [
      '#type' => 'key_select',
      '#key_filters' => ['type' => 'duo'],
      '#title' => t('Duo key'),
      '#default_value' => !empty($this->settings['duo_key']) ? $this->settings['duo_key'] : '',
      '#description' => t('The Duo key with all the credentials from the Duo administrative interface'),
      '#required' => TRUE,
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array $form, FormStateInterface $form_state) {
    // Invoke the verification method again with the form state values.
    $valid = $this->duoVerifiedLogin($form_state->getValue('duo_integration'), $form_state->getValue('duo_secret'), $form_state->getValue('duo_application'), $form_state->getValue('sig_response'));
    return !empty($valid);
  }

  /**
   * {@inheritdoc}
   */
  public function getFallbacks() {
    return ($this->pluginDefinition['fallbacks']) ?: '';
  }

  /**
   * {@inheritdoc}
   */
  public function isFallback() {
    return ($this->pluginDefinition['isFallback']) ?: FALSE;
  }

  /**
   * Helper method to handle the Duo iframe POST.
   *
   * @return mixed
   *   User name if logged in or false.
   */
  public function duoVerifiedLogin($ikey, $skey, $akey, $sig_resp) {
    $name = FALSE;

    if (empty($sig_resp)) {
      $this->messenger()->addError("The required information wasn't received. Please try again.");
      return FALSE;
    }

    $name = $this->duo->verifyResponse($ikey, $skey, $akey, $sig_resp);
    $this->responseErrors($name);
    if (!$name || !user_load_by_name($name)) {
      $this->messenger()->addError("Secondary authentication credentials were invalid, or the user cannot be loaded.");
      return FALSE;
    }

    return $name;
  }

  /**
   * Helper function to deal with errors from Duo API calls.
   *
   * @param string $response
   *   The response string from Duo.
   * @param bool $verbose
   *   Whether or not to log the error message.
   *
   * @return bool
   *   Whether or not there were errors.
   */
  public function responseErrors($response, $verbose = TRUE) {
    $has_errors = FALSE;
    if (strpos("|", $response) !== FALSE) {
      [$response_code, $message] = explode('|', $response);
      $has_errors = ($response_code == 'ERR');
      if ($has_errors && $verbose) {
        $this->messenger()->addError(HTML::escape($message));
      }
    }
    return $has_errors;
  }

}
